@@ -4,20 +4,39 @@ use warnings;
 use Data::Dumper;
 
 use Getopt::Long qw(GetOptions);
+use Encode qw(encode_utf8 decode_utf8);
 our @constants;
+no warnings 'recursion';
 BEGIN {
-    my $err;
-    eval '
-        use Sereal::Encoder::Constants qw(:all);
-        @constants= @Sereal::Encoder::Constants::EXPORT_OK;
-        print "Loaded constants from $INC{q(Sereal/Encoder/Constants.pm)}\n";
-        1;
-    ' or do { $err= $@; eval '
-        use Sereal::Decoder::Constants qw(:all);
-        @constants= @Sereal::Decoder::Constants::EXPORT_OK;
-        print "Loaded constants from $INC{q(Sereal/Decoder/Constants.pm)}\n";
-        1;
-    ' } or die "No encoder/decoder constants: $err\n$@";
+    my $add_use_blib= "";
+    my $use= "";
+    my @check;
+    for my $type ("Decoder","Encoder") {
+        if (-e "blib/lib/Sereal/$type/Constants.pm") {
+            $add_use_blib="use blib;";
+            @check= ($type);
+            last;
+        }
+        push @check, $type;
+    }
+
+    my @err;
+    foreach my $check (@check) {
+        if (eval(my $code= sprintf '
+                %s
+                use Sereal::%s::Constants qw(:all);
+                @constants= @Sereal::%s::Constants::EXPORT_OK;
+                print "Loaded constants from $INC{q(Sereal/%s/Constants.pm)}\n";
+                1;
+            ', $add_use_blib, ($check) x 3))
+        {
+            @err= ();
+            last;
+        } else {
+            push @err, "Error:",$@ || "Zombie Error","\nCode:\n$code";
+        }
+    }
+    die @err if @err;
 }
 
 my $done;
@@ -25,12 +44,18 @@ my $data;
 my $hlen;
 my $indent = "";
 
+sub _chop_data_prefix {
+    my ($len)= @_;
+    die "Unexpected end of packet" unless length($data) >= $len;
+    return substr($data,0,$len,'');
+}
+
 sub parse_header {
   $data =~ s/^(=[s\xF3]rl)(.)// or die "invalid header: $data";
   $done .= $1 . $2;
   my $flags = $2;
   my $len = varint();
-  my $hdr = substr($data, 0, $len, '');
+  my $hdr = _chop_data_prefix( $len );
 
   my $proto_version = ord($flags) & SRL_PROTOCOL_VERSION_MASK;
   print "Sereal protocol version: $proto_version\n";
@@ -81,20 +106,20 @@ sub parse_header {
 my ($len_f, $len_d, $len_D);
 sub parse_float {
     $len_f||= length(pack("f",0));
-    my $v= substr($data,0,$len_f,"");
+    my $v= _chop_data_prefix( $len_f );
     $done .= $v;
     return unpack("f",$v);
 }
 sub parse_double {
     $len_d||= length(pack("d",0));
-    my $v= substr($data,0,$len_d,"");
+    my $v= _chop_data_prefix( $len_d );
     $done .= $v;
     return unpack("d",$v);
 }
 sub parse_long_double {
     $len_D ||= eval { length(pack("D",0.0)) };
     die "Long double not supported" unless $len_D;
-    my $v= substr($data, 0, $len_D, "");
+    my $v= _chop_data_prefix( $len_D );
     $done .= $v;
     return unpack("D",$v);
 }
@@ -106,7 +131,7 @@ sub parse_sv {
   my ($ind) = @_;
 
   my $p= length($done);
-  my $t = substr($data, 0, 1, '');
+  my $t = _chop_data_prefix( 1 );
   $done .= $t;
   my $o = ord($t);
   my $bv= $o;
@@ -130,15 +155,16 @@ sub parse_sv {
   elsif ($o >= SRL_HDR_SHORT_BINARY_LOW) {
     $o -= SRL_HDR_SHORT_BINARY_LOW;
     my $len = $o;
-    my $str = substr($data, 0, $len, '');
+    my $str = _chop_data_prefix( $len );
     $done .= $str;
-    printf "SHORT_BINARY(%u): '%s'\n", $len, $str;
+    printf "SHORT_BINARY(%u): '%s' (%s)\n", $len, encode_utf8($str), unpack("H*", $str);
   }
   elsif ($o == SRL_HDR_BINARY || $o == SRL_HDR_STR_UTF8) {
     my $l = varint();
-    my $str = substr($data, 0, $l, ""); # fixme UTF8
+    my $str = _chop_data_prefix( $l ); # fixme UTF8
     $done .= $str;
-    printf( ($o == SRL_HDR_STR_UTF8 ? "STR_UTF8" : "BINARY")."(%u): '%s'\n", $l, $str);
+    $str= decode_utf8($str) if $o == SRL_HDR_STR_UTF8;
+    printf( ($o == SRL_HDR_STR_UTF8 ? "STR_UTF8" : "BINARY")."(%u): '%s' (%s)\n", $l, encode_utf8($str), unpack("H*", encode_utf8($str)));
   }
   elsif ($o == SRL_HDR_FLOAT) {
     printf "FLOAT(%f)\n", parse_float();
@@ -246,22 +272,19 @@ sub parse_av {
   printf "(%u)\n", $len;
   $ind .= "  ";
   while ($len--) {
-    my $t = substr($data, 0, 1);
-    my $o = ord($t);
-      parse_sv($ind);
+    parse_sv($ind,\$len);
   }
 }
 
 sub parse_hv {
   my ($ind, $o) = @_;
-  my $len = (defined $o ? $o & 15 : varint()) * 2;
+  my $len = (defined $o ? $o & 15 : varint());
   printf "(%u)\n", $len;
   $ind .= "  ";
-  my $flipflop = 0;
   while ($len--) {
-    my $t = substr($data, 0, 1);
-    my $o = ord($t);
-    printf  "$fmt2%s:\n",("") x $lead_items, $ind, ($flipflop++ %2 == 1 ? "VALUE" : "KEY");
+    printf  "$fmt2%s:\n",("") x $lead_items, $ind, "KEY";
+    parse_sv($ind."  ");
+    printf  "$fmt2%s:\n",("") x $lead_items, $ind,  "VALUE";
     parse_sv($ind."  ");
   }
 }
@@ -272,13 +295,13 @@ sub varint {
   my $x = 0;
   my $lshift = 0;
   while (length($data) && ord(substr($data, 0, 1)) & 0x80) {
-    my $c = ord(substr($data, 0, 1, ''));
+    my $c = ord(_chop_data_prefix( 1 ));
     $done .= chr($c);
     $x += ($c & 0x7F) << $lshift;
     $lshift += 7;
   }
   if (length($data)) {
-    my $c = ord(substr($data, 0, 1, ''));
+    my $c = ord(_chop_data_prefix( 1 ));
     $done .= chr($c);
     $x += $c << $lshift;
   }
@@ -288,12 +311,12 @@ sub varint {
   return $x;
 }
 
-BEGIN{
-my $_shift= length(pack"j",0) * 8 - 1;
-sub zigzag {
-    my $n= varint();
-    return ($n >> 1) ^ (-($n & 1));
+sub _zigzag {
+    my $n= $_[0];
+    return $n & 1 ? -(($n >> 1)+1) : ($n >> 1);
 }
+sub zigzag {
+    return _zigzag(varint());
 }
 
 GetOptions(
